#!/usr/bin/env python
""" Convolve ASKAP images to common resolution """
__author__ = "Alec Thomson"

import logging
import os
import sys
from functools import partial
from hashlib import new
from typing import Dict, List, Optional, Tuple

import astropy.wcs
import numpy as np
import schwimmbad
from astropy import units as u
from astropy.io import ascii, fits
from astropy.table import Table
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales
from radio_beam import Beam, Beams
from radio_beam.utils import BeamError
from tqdm import tqdm

from racs_tools import au2
from racs_tools.convolve_uv import smooth
from racs_tools.logging import logger, setup_logger

#############################################
#### ADAPTED FROM SCRIPT BY T. VERNSTROM ####
#############################################


def round_up(n: float, decimals: int = 0) -> float:
    """Round to number of decimals

    Args:
        n (float): Number to round.
        decimals (int, optional): Number of decimals. Defaults to 0.

    Returns:
        float: Rounded number.
    """
    multiplier = 10**decimals
    return np.ceil(n * multiplier) / multiplier


def my_ceil(a: float, precision: float = 0.0) -> float:
    """Modified ceil function to round up to precision

    Args:
        a (float): Number to round.
        precision (float, optional): Precision of rounding. Defaults to 0.

    Returns:
        float: Rounded number.
    """
    return np.round(a + 0.5 * 10 ** (-precision), precision)


def getbeam(
    old_beam: Beam,
    new_beam: Beam,
    dx: u.Quantity,
    dy: u.Quantity,
    cutoff: Optional[float] = None,
) -> Tuple[Beam, float]:
    """Get the beam to use for smoothing

    Args:
        old_beam (Beam): Current beam.
        new_beam (Beam): Target beam.
        dx (u.Quantity): Pixel size in x.
        dy (u.Quantity): Pixel size in y.
        cutoff (float, optional): Cutoff for beamsize in arcsec. Defaults to None.

    Raises:
        err: If beam deconvolution fails.

    Returns:
        Tuple[Beam, float]: Convolving beam and scaling factor.
    """
    logger.info(f"Current beam is {old_beam!r}")

    if cutoff is not None and old_beam.major.to(u.arcsec) > cutoff * u.arcsec:
        return (
            Beam(
                major=np.nan * u.deg,
                minor=np.nan * u.deg,
                pa=np.nan * u.deg,
            ),
            np.nan,
        )

    if new_beam == old_beam:
        conbm = Beam(
            major=0 * u.deg,
            minor=0 * u.deg,
            pa=0 * u.deg,
        )
        fac = 1.0
        logger.warning(
            f"New beam {new_beam!r} and old beam {old_beam!r} are the same. Won't attempt convolution."
        )
        return conbm, fac
    try:
        conbm = new_beam.deconvolve(old_beam)
    except Exception as err:
        logger.warning(f"Could not deconvolve. New: {new_beam!r}, Old: {old_beam!r}")
        raise err
    fac, amp, outbmaj, outbmin, outbpa = au2.gauss_factor(
        beamConv=[
            conbm.major.to(u.arcsec).value,
            conbm.minor.to(u.arcsec).value,
            conbm.pa.to(u.deg).value,
        ],
        beamOrig=[
            old_beam.major.to(u.arcsec).value,
            old_beam.minor.to(u.arcsec).value,
            old_beam.pa.to(u.deg).value,
        ],
        dx1=dx.to(u.arcsec).value,
        dy1=dy.to(u.arcsec).value,
    )

    return conbm, fac


def getimdata(cubenm: str) -> dict:
    """Get image data from FITS file

    Args:
        cubenm (str): File name of image.

    Returns:
        dict: Data and metadata.
    """
    logger.info(f"Getting image data from {cubenm}")
    with fits.open(cubenm, memmap=True, mode="denywrite") as hdu:
        w = astropy.wcs.WCS(hdu[0])
        pixelscales = proj_plane_pixel_scales(w)

        dxas = pixelscales[0] * u.deg
        dyas = pixelscales[1] * u.deg

        if len(hdu[0].data.shape) == 4:
            # has spectral, polarization axes
            data = hdu[0].data[0, 0]
        else:
            data = hdu[0].data
        nx, ny = data.shape[-1], data.shape[-2]

        old_beam = Beam.from_fits_header(hdu[0].header)

        datadict = {
            "filename": os.path.basename(cubenm),
            "image": data,
            "4d": (len(hdu[0].data.shape) == 4),
            "header": hdu[0].header,
            "old_beam": old_beam,
            "nx": nx,
            "ny": ny,
            "dx": dxas,
            "dy": dyas,
        }
    return datadict


def savefile(
    newimage: np.ndarray,
    filename: str,
    header: fits.Header,
    final_beam: Beam,
    outdir: str = ".",
) -> None:
    """Save smoothed image to FITS file

    Args:
        newimage (np.ndarray): Smoothed image.
        filename (str): File name.
        header (fits.Header): FITS header.
        final_beam (Beam): New beam.
        outdir (str, optional): Output directory. Defaults to ".".
    """
    outfile = f"{outdir}/{filename}"
    logger.info(f"Saving to {outfile}")
    beam = final_beam
    header = beam.attach_to_header(header)
    fits.writeto(outfile, newimage.astype(np.float32), header=header, overwrite=True)


def worker(
    file: str,
    outdir: str,
    new_beam: Beam,
    conv_mode: str,
    suffix: str = "",
    prefix: str = "",
    cutoff: Optional[float] = None,
    dryrun: bool = False,
) -> dict:
    """Parallel worker function

    Args:
        file (str): FITS file to smooth.
        outdir (str): Output directory.
        new_beam (Beam): Target PSF.
        conv_mode (str): Convolving mode.
        suffix (str, optional): Filename suffix. Defaults to "".
        prefix (str, optional): Filename prefix. Defaults to "".
        cutoff (float, optional): PSF cutoff. Defaults to None.
        dryrun (bool, optional): Do a dryrun. Defaults to False.

    Returns:
        dict: Output data.
    """
    logger.info(f"Working on {file}")

    if outdir is None:
        outdir = os.path.dirname(file)

    if outdir == "":
        outdir = "."

    outfile = os.path.basename(file)
    outfile = outfile.replace(".fits", "") + f".{suffix}.fits"
    if prefix is not None:
        outfile = prefix + outfile
    datadict = getimdata(file)

    conbeam, sfactor = getbeam(
        old_beam=datadict["old_beam"],
        new_beam=new_beam,
        dx=datadict["dx"],
        dy=datadict["dy"],
        cutoff=cutoff,
    )

    datadict.update({"conbeam": conbeam, "final_beam": new_beam, "sfactor": sfactor})
    if not dryrun:
        if np.isnan(sfactor) or np.isnan(conbeam):
            logger.warning(f"Setting {outfile} to NaN")
            newim = datadict["image"] * np.nan
        elif (
            conbeam == Beam(major=0 * u.deg, minor=0 * u.deg, pa=0 * u.deg)
            and sfactor == 1
        ):
            newim = datadict["image"]
        else:
            newim = smooth(
                image=datadict["image"],
                old_beam=datadict["old_beam"],
                final_beam=datadict["final_beam"],
                dx=datadict["dx"],
                dy=datadict["dy"],
                sfactor=datadict["sfactor"],
                conbeam=datadict["conbeam"],
                conv_mode=conv_mode,
            )
        if datadict["4d"]:
            # make it back into a 4D image
            newim = np.expand_dims(np.expand_dims(newim, axis=0), axis=0)
        datadict.update(
            {
                "newimage": newim,
            }
        )
        savefile(
            newimage=datadict["newimage"],
            filename=outfile,
            header=datadict["header"],
            final_beam=datadict["final_beam"],
            outdir=outdir,
        )

    return datadict


def getmaxbeam(
    files: List[str],
    conv_mode: str = "robust",
    target_beam: Beam = None,
    cutoff: Optional[float] = None,
    tolerance: float = 0.0001,
    nsamps: float = 200,
    epsilon: float = 0.0005,
) -> Tuple[Beam, Beams]:
    """Get the smallest common beam.

    Args:
        files (List[str]): List of FITS files.
        conv_mode (str, optional): Convolution mode. Defaults to "robust".
        target_beam (Beam, optional): Target PSF. Defaults to None.
        cutoff (float, optional): Cutoff PSF BMAJ in arcsec. Defaults to None.
        tolerance (float, optional): Common beam tolerance. Defaults to 0.0001.
        nsamps (float, optional): Common beam samples. Defaults to 200.
        epsilon (float, optional): Commonn beam epsilon. Defaults to 0.0005.

    Raises:
        Exception: X and Y pixel sizes are not the same.
        Exception: Convolving beam will be undersampled on pixel grid.

    Returns:
        Tuple[Beam, Beams]: Common beam and all beams.
    """
    beams_list = []
    for file in files:
        header = fits.getheader(file, memmap=True)
        beam = Beam.from_fits_header(header)
        beams_list.append(beam)

    beams = Beams(
        [beam.major.to(u.deg).value for beam in beams_list] * u.deg,
        [beam.minor.to(u.deg).value for beam in beams_list] * u.deg,
        [beam.pa.to(u.deg).value for beam in beams_list] * u.deg,
    )
    if cutoff is not None:
        flags = beams.major > cutoff * u.arcsec
        if np.all(flags):
            logger.critical(
                "All beams are larger than cutoff. All outputs will be blanked!"
            )
            nan_beam = Beam(np.nan * u.deg, np.nan * u.deg, np.nan * u.deg)
            nan_beams = Beams(
                [np.nan for beam in beams_list] * u.deg,
                [np.nan for beam in beams_list] * u.deg,
                [np.nan for beam in beams_list] * u.deg,
            )
            return nan_beam, nan_beams
    else:
        flags = np.array([False for beam in beams])

    if target_beam is None:
        # Find the common beam
        try:
            cmn_beam = beams[~flags].common_beam(
                tolerance=tolerance, epsilon=epsilon, nsamps=nsamps
            )
        except BeamError:
            logger.warning(
                "Couldn't find common beam with defaults\nTrying again with smaller tolerance"
            )
            cmn_beam = beams[~flags].common_beam(
                tolerance=tolerance * 0.1, epsilon=epsilon, nsamps=nsamps
            )

        # Round up values
        cmn_beam = Beam(
            major=my_ceil(cmn_beam.major.to(u.arcsec).value, precision=1) * u.arcsec,
            minor=my_ceil(cmn_beam.minor.to(u.arcsec).value, precision=1) * u.arcsec,
            pa=round_up(cmn_beam.pa.to(u.deg), decimals=2),
        )

    else:
        cmn_beam = target_beam

    target_header = header
    w = WCS(target_header)
    pixelscales = proj_plane_pixel_scales(w)

    dx = pixelscales[0] * u.deg
    dy = pixelscales[1] * u.deg
    if not np.isclose(dx, dy):
        raise Exception("GRID MUST BE SAME IN X AND Y")
    grid = dy
    if conv_mode != "robust":
        # Get the minor axis of the convolving beams
        minorcons = []
        for beam in beams:
            minorcons += [cmn_beam.deconvolve(beam).minor.to(u.arcsec).value]
        minorcons = np.array(minorcons) * u.arcsec
        samps = minorcons / grid.to(u.arcsec)
        # Check that convolving beam will be Nyquist sampled
        if any(samps.value < 2):
            # Set the convolving beam to be Nyquist sampled
            nyq_con_beam = Beam(major=grid * 2, minor=grid * 2, pa=0 * u.deg)
            # Find new target based on common beam * Nyquist beam
            # Not sure if this is best - but it works
            nyq_beam = cmn_beam.convolve(nyq_con_beam)
            nyq_beam = Beam(
                major=my_ceil(nyq_beam.major.to(u.arcsec).value, precision=1)
                * u.arcsec,
                minor=my_ceil(nyq_beam.minor.to(u.arcsec).value, precision=1)
                * u.arcsec,
                pa=round_up(nyq_beam.pa.to(u.deg), decimals=2),
            )
            logger.info(f"Smallest common Nyquist sampled beam is: {nyq_beam!r}")
            if target_beam is not None:
                if target_beam < nyq_beam:
                    logger.warning("TARGET BEAM WILL BE UNDERSAMPLED!")
                    raise Exception("CAN'T UNDERSAMPLE BEAM - EXITING")
            else:
                logger.warning("COMMON BEAM WILL BE UNDERSAMPLED!")
                logger.warning("SETTING COMMON BEAM TO NYQUIST BEAM")
                cmn_beam = nyq_beam

    return cmn_beam, beams


def writelog(output: List[Dict], commonbeam_log: str):
    """Write beamlog file.

    Args:
        output (List[Dict]): List of dictionaries containing output.
        commonbeam_log (str): Name of log file.
    """
    commonbeam_tab = Table()
    commonbeam_tab.add_column([out["filename"] for out in output], name="FileName")
    # Origina
    commonbeam_tab.add_column(
        [out["old_beam"].major.to(u.deg).value for out in output] * u.deg,
        name="Original BMAJ",
    )
    commonbeam_tab.add_column(
        [out["old_beam"].minor.to(u.deg).value for out in output] * u.deg,
        name="Original BMIN",
    )
    commonbeam_tab.add_column(
        [out["old_beam"].pa.to(u.deg).value for out in output] * u.deg,
        name="Original BPA",
    )
    # Target
    commonbeam_tab.add_column(
        [out["final_beam"].major.to(u.deg).value for out in output] * u.deg,
        name="Target BMAJ",
    )
    commonbeam_tab.add_column(
        [out["final_beam"].minor.to(u.deg).value for out in output] * u.deg,
        name="Target BMIN",
    )
    commonbeam_tab.add_column(
        [out["final_beam"].pa.to(u.deg).value for out in output] * u.deg,
        name="Target BPA",
    )
    # Convolving
    commonbeam_tab.add_column(
        [out["conbeam"].major.to(u.deg).value for out in output] * u.deg,
        name="Convolving BMAJ",
    )
    commonbeam_tab.add_column(
        [out["conbeam"].minor.to(u.deg).value for out in output] * u.deg,
        name="Convolving BMIN",
    )
    commonbeam_tab.add_column(
        [out["conbeam"].pa.to(u.deg).value for out in output] * u.deg,
        name="Convolving BPA",
    )

    # Write to log file
    units = ""
    for col in commonbeam_tab.colnames:
        unit = commonbeam_tab[col].unit
        unit = str(unit)
        units += unit + " "
    commonbeam_tab.meta["comments"] = [units]
    ascii.write(
        commonbeam_tab, output=commonbeam_log, format="commented_header", overwrite=True
    )
    logger.info(f"Convolving log written to {commonbeam_log}")


def main(
    pool,
    infile: list = [],
    prefix: Optional[str] = None,
    suffix: Optional[str] = None,
    outdir: Optional[str] = None,
    conv_mode: str = "robust",
    dryrun: bool = False,
    bmaj: Optional[float] = None,
    bmin: Optional[float] = None,
    bpa: Optional[float] = None,
    log: Optional[str] = None,
    circularise: bool = False,
    cutoff: Optional[float] = None,
    tolerance: float = 0.0001,
    nsamps: int = 200,
    epsilon: float = 0.0005,
):
    """Main script.

    Args:
        pool (mp.Pool): Multiprocessing pool.
        infile (list, optional): List of images to convolve. Defaults to [].
        prefix (str, optional): Output prefix. Defaults to None.
        suffix (str, optional): Output suffix. Defaults to None.
        outdir (str, optional): Output directory. Defaults to None.
        conv_mode (str, optional): Colvolution mode. Defaults to "robust".
        dryrun (bool, optional): Do a dryrun. Defaults to False.
        bmaj (float, optional): Target BMAJ. Defaults to None.
        bmin (float, optional): Target BMIN. Defaults to None.
        bpa (float, optional): Target BPA. Defaults to None.
        log (str, optional): Input beamlog. Defaults to None.
        circularise (bool, optional): Make beam circular. Defaults to False.
        cutoff (float, optional): Cutoff beams. Defaults to None.
        tolerance (float, optional): Common tolerance. Defaults to 0.0001.
        nsamps (int, optional): Common samples. Defaults to 200.
        epsilon (float, optional): Common epsilon. Defaults to 0.0005.


    Raises:
        Exception: If no files are found.
        Exception: If invalid convolution mode is specified.
        Exception: If partial target beam is specified.
        Exception: If target beam cannot be used.
    """

    if dryrun:
        logger.info("Doing a dry run -- no files will be saved")
    # Fix up outdir
    if outdir is not None:
        outdir = os.path.abspath(outdir)

    # Get file list
    files = sorted(infile)
    if files == []:
        raise Exception("No files found!")

    # Parse args
    logger.info(f"Convolution mode: {conv_mode}")
    if not conv_mode in ["robust", "scipy", "astropy", "astropy_fft"]:
        raise Exception("Please select valid convolution method!")

    logger.info(f"Using convolution method {conv_mode}")
    if conv_mode == "robust":
        logger.info("This is the most robust method. And fast!")
    elif conv_mode == "scipy":
        logger.info("This fast, but not robust to NaNs or small PSF changes")
    else:
        logger.info("This is slower, but robust to NaNs, but not to small PSF changes")

    nonetest = [test is None for test in [bmaj, bmin, bpa]]

    if all(nonetest):
        target_beam = None

    elif not all(nonetest) and any(nonetest):
        raise Exception("Please specify all target beam params!")

    elif not all(nonetest) and not any(nonetest):
        target_beam = Beam(bmaj * u.arcsec, bmin * u.arcsec, bpa * u.deg)
        logger.info(f"Target beam is {target_beam!r}")

    # Find smallest common beam
    big_beam, allbeams = getmaxbeam(
        files,
        conv_mode=conv_mode,
        target_beam=target_beam,
        cutoff=cutoff,
        tolerance=tolerance,
        nsamps=nsamps,
        epsilon=epsilon,
    )

    if target_beam is not None:
        logger.info("Checking that target beam will deconvolve...")

        mask_count = 0
        failed = []
        for i, (beam, file) in enumerate(
            tqdm(
                zip(allbeams, files),
                total=len(allbeams),
                desc="Deconvolving",
                disable=(logger.level > logging.INFO),
            )
        ):
            if cutoff is not None and beam.major.to(u.arcsec) > cutoff * u.arcsec:
                continue
            try:
                target_beam.deconvolve(beam)
            except ValueError:
                mask_count += 1
                failed.append(file)
            except BeamError as be:
                # BeamError should not be raised if beams are equal
                if target_beam != beam:
                    raise BeamError(be)
        if mask_count > 0:
            logger.warning("The following images could not reach target resolution:")
            logger.warning(failed)

            raise Exception("Please choose a larger target beam!")

        else:
            new_beam = target_beam

    else:
        new_beam = big_beam

    if circularise:
        logger.info("Circular beam requested, setting BMIN=BMAJ and BPA=0")
        new_beam = Beam(
            major=new_beam.major,
            minor=new_beam.major,
            pa=0 * u.deg,
        )

    logger.info(f"Final beam is {new_beam!r}")

    output = list(
        pool.map(
            partial(
                worker,
                outdir=outdir,
                new_beam=new_beam,
                conv_mode=conv_mode,
                suffix=suffix,
                prefix=prefix,
                cutoff=cutoff,
                dryrun=dryrun,
            ),
            files,
        )
    )

    if log is not None:
        writelog(output, log)

    logger.info("Done!")
    return new_beam


def cli():
    """Command-line interface"""
    import argparse

    # Help string to be shown using the -h option
    descStr = """
    Smooth a field of 2D images to a common resolution.

    - Parallelisation can run using multiprocessing or MPI.

    - Default names of output files are /path/to/beamlog{infile//.fits/.{SUFFIX}.fits}

    - By default, the smallest common beam will be automatically computed.
    - Optionally, you can specify a target beam to use.

    """

    # Parse the command line options
    parser = argparse.ArgumentParser(
        description=descStr, formatter_class=argparse.RawTextHelpFormatter
    )

    parser.add_argument(
        "infile",
        metavar="infile",
        type=str,
        help="Input FITS image(s) to smooth (can be a wildcard) - beam info must be in header.",
        nargs="+",
    )

    parser.add_argument(
        "-p",
        "--prefix",
        dest="prefix",
        type=str,
        default=None,
        help="Add prefix to output filenames.",
    )

    parser.add_argument(
        "-s",
        "--suffix",
        dest="suffix",
        type=str,
        default="sm",
        help="Add suffix to output filenames [sm].",
    )

    parser.add_argument(
        "-o",
        "--outdir",
        dest="outdir",
        type=str,
        default=None,
        help="Output directory of smoothed FITS image(s) [same as input file].",
    )

    parser.add_argument(
        "--conv_mode",
        dest="conv_mode",
        type=str,
        choices=["robust", "scipy", "astropy", "astropy_fft"],
        default="robust",
        help="""Which method to use for convolution [robust].
        'robust' computes the analytic FT of the convolving Gaussian.
        Note this mode can now handle NaNs in the data.
        Can also be 'scipy', 'astropy', or 'astropy_fft'.
        Note these other methods cannot cope well with small convolving beams.
        """,
    )

    parser.add_argument(
        "-v", "--verbosity", default=0, action="count", help="Increase output verbosity"
    )

    parser.add_argument(
        "-d",
        "--dryrun",
        dest="dryrun",
        action="store_true",
        help="Compute common beam and stop [False].",
    )

    parser.add_argument(
        "--bmaj",
        dest="bmaj",
        type=float,
        default=None,
        help="Target BMAJ (arcsec) to convolve to [None].",
    )

    parser.add_argument(
        "--bmin",
        dest="bmin",
        type=float,
        default=None,
        help="Target BMIN (arcsec) to convolve to [None].",
    )

    parser.add_argument(
        "--bpa",
        dest="bpa",
        type=float,
        default=None,
        help="Target BPA (deg) to convolve to [None].",
    )

    parser.add_argument(
        "--log",
        dest="log",
        type=str,
        default=None,
        help="Name of beamlog file. If provided, save beamlog data to a file [None - not saved].",
    )

    parser.add_argument(
        "--logfile",
        default=None,
        type=str,
        help="Save logging output to file",
    )

    parser.add_argument(
        "-c",
        "--cutoff",
        dest="cutoff",
        type=float,
        default=None,
        help="Cutoff BMAJ value (arcsec) -- Blank channels with BMAJ larger than this [None -- no limit]",
    )

    parser.add_argument(
        "--circularise",
        action="store_true",
        help="Circularise the final PSF -- Sets the BMIN = BMAJ, and BPA=0.",
    )

    parser.add_argument(
        "-t",
        "--tolerance",
        dest="tolerance",
        type=float,
        default=0.0001,
        help="tolerance for radio_beam.commonbeam.",
    )

    parser.add_argument(
        "-e",
        "--epsilon",
        dest="epsilon",
        type=float,
        default=0.0005,
        help="epsilon for radio_beam.commonbeam.",
    )

    parser.add_argument(
        "-n",
        "--nsamps",
        dest="nsamps",
        type=int,
        default=200,
        help="nsamps for radio_beam.commonbeam.",
    )

    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        "--ncores",
        dest="n_cores",
        default=1,
        type=int,
        help="Number of processes (uses multiprocessing).",
    )
    group.add_argument(
        "--mpi", dest="mpi", default=False, action="store_true", help="Run with MPI."
    )

    args = parser.parse_args()
    # Set up logging
    logger = setup_logger(
        verbosity=args.verbosity,
        filename=args.logfile,
    )

    pool = schwimmbad.choose_pool(mpi=args.mpi, processes=args.n_cores)
    if args.mpi:
        if not pool.is_master():
            pool.wait()
            sys.exit(0)

    arg_dict = vars(args)
    # pop unwanted arguments
    _ = arg_dict.pop("mpi")
    _ = arg_dict.pop("n_cores")
    _ = arg_dict.pop("verbosity")
    _ = arg_dict.pop("logfile")

    _ = main(pool, **arg_dict)
    pool.close()


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
from typing import Optional

try:
    from mpi4py import MPI

    myPE = MPI.COMM_WORLD.Get_rank()
except ImportError:
    myPE = 0

logging.captureWarnings(True)
logger = logging.getLogger("racs_tools")
logger.setLevel(logging.WARNING)

formatter = logging.Formatter(
    fmt=f"[{myPE}] %(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


def setup_logger(
    verbosity: int = 0,
    filename: Optional[str] = None,
):
    if verbosity == 0:
        level = logging.WARNING
    elif verbosity == 1:
        level = logging.INFO
    elif verbosity >= 2:
        level = logging.DEBUG

    ch = logging.StreamHandler()
    ch.setLevel(level)
    ch.setFormatter(formatter)

    logger.setLevel(level)
    logger.addHandler(ch)

    if filename is not None:
        fh = logging.FileHandler(filename)
        fh.setLevel(level)
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    return logger



#!/usr/bin/env python3
""" Convolve ASKAP cubes to common resolution """
__author__ = "Alec Thomson"

import logging
import os
import stat
import sys
import warnings
from typing import Dict, List, Tuple

import numpy as np
import scipy.signal
from astropy import units as u
from astropy.io import ascii, fits
from astropy.table import Table
from astropy.utils.exceptions import AstropyWarning
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales
from radio_beam import Beam, Beams
from radio_beam.utils import BeamError
from spectral_cube import SpectralCube
from spectral_cube.utils import SpectralCubeWarning
from tqdm import tqdm, trange

from racs_tools import au2
from racs_tools.beamcon_2D import my_ceil, round_up
from racs_tools.convolve_uv import smooth
from racs_tools.logging import logger, setup_logger

mpiSwitch = False
if (
    os.environ.get("OMPI_COMM_WORLD_SIZE") is not None
    or int(os.environ.get("SLURM_NTASKS") or 1) > 1
):
    mpiSwitch = True

if mpiSwitch:
    try:
        from mpi4py import MPI
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "Script called with mpiexec/mpirun/srun, but mpi4py not installed"
        )
    # Get the processing environment
    comm = MPI.COMM_WORLD
    nPE = comm.Get_size()
    myPE = comm.Get_rank()
else:
    nPE = 1
    myPE = 0

warnings.filterwarnings(action="ignore", category=SpectralCubeWarning, append=True)
warnings.simplefilter("ignore", category=AstropyWarning)
warnings.filterwarnings("ignore", message="invalid value encountered in true_divide")

#############################################
#### ADAPTED FROM SCRIPT BY T. VERNSTROM ####
#############################################


class Error(OSError):
    pass


class SameFileError(Error):
    """Raised when source and destination are the same file."""


class SpecialFileError(OSError):
    """Raised when trying to do a kind of operation (e.g. copying) which is
    not supported on a special file (e.g. a named pipe)"""


class ExecError(OSError):
    """Raised when a command could not be executed"""


class ReadError(OSError):
    """Raised when an archive cannot be read"""


class RegistryError(Exception):
    """Raised when a registry operation with the archiving
    and unpacking registeries fails"""


def _samefile(src, dst):
    # Macintosh, Unix.
    if hasattr(os.path, "samefile"):
        try:
            return os.path.samefile(src, dst)
        except OSError:
            return False


def copyfile(src, dst, *, follow_symlinks=True):
    """Copy data from src to dst.

    If follow_symlinks is not set and src is a symbolic link, a new
    symlink will be created instead of copying the file it points to.

    """
    if _samefile(src, dst):
        raise SameFileError(f"{src!r} and {dst!r} are the same file")

    for fn in [src, dst]:
        try:
            st = os.stat(fn)
        except OSError:
            # File most likely does not exist
            pass
        else:
            # XXX What about other special files? (sockets, devices...)
            if stat.S_ISFIFO(st.st_mode):
                raise SpecialFileError("`%s` is a named pipe" % fn)

    if not follow_symlinks and os.path.islink(src):
        os.symlink(os.readlink(src), dst)
    else:
        with open(src, "rb") as fsrc:
            with open(dst, "wb") as fdst:
                copyfileobj(fsrc, fdst)
    return dst


def copyfileobj(fsrc, fdst, length=16 * 1024):
    # copied = 0
    total = os.fstat(fsrc.fileno()).st_size
    with tqdm(
        total=total,
        disable=(logger.level > logging.INFO),
        unit_scale=True,
        desc="Copying file",
    ) as pbar:
        while True:
            buf = fsrc.read(length)
            if not buf:
                break
            fdst.write(buf)
            copied = len(buf)
            pbar.update(copied)


def getbeams(file: str, header: fits.Header) -> Tuple[Table, int, str]:
    """Get beam information from a fits file or beamlog.

    Args:
        file (str): FITS filename.
        header (fits.Header): FITS header.

    Returns:
        Tuple[Table, int, str]: Table of beams, number of beams, and beamlog filename.
    """
    # Add beamlog info to dict just in case
    dirname = os.path.dirname(file)
    basename = os.path.basename(file)
    if dirname == "":
        dirname = "."
    beamlog = f"{dirname}/beamlog.{basename}".replace(".fits", ".txt")

    # First check for CASA beams
    try:
        headcheck = header["CASAMBM"]
    except KeyError:
        headcheck = False
    if headcheck:
        logger.info(
            "CASA beamtable found in header - will use this table for beam calculations"
        )
        with fits.open(file) as hdul:
            hdu = hdul.pop("BEAMS")
            beams = Table.read(hdu)

    # Otherwise use beamlog file
    else:
        try:
            logger.info("No CASA beamtable found in header - looking for beamlogs")
            logger.info(f"Getting beams from {beamlog}")

            beams = Table.read(beamlog, format="ascii.commented_header")
            # Header looks like:
            # colnames=['Channel', 'BMAJarcsec', 'BMINarcsec', 'BPAdeg']
            # But needs some fix up - astropy can't read the header properly
            for col in beams.colnames:
                idx = col.find("[")
                if idx == -1:
                    new_col = col
                    unit = u.Unit("")
                else:
                    new_col = col[:idx]
                    unit = u.Unit(col[idx + 1 : -1])
                beams[col].unit = unit
                beams[col].name = new_col
        except FileNotFoundError:
            logger.warning("No beamlog found")
            logger.warning("Using header beam - assuming a constant beam")
            try:
                beam: Beam = Beam.from_fits_header(header)
                wcs = WCS(header)
                spec_axis = wcs.spectral
                _nchan = spec_axis.array_shape[0]
                beams = Table()
                beams.add_column(np.arange(_nchan), name="Channel")
                beams.add_column([beam.major.to(u.arcsec)] * _nchan, name="BMAJ")
                beams.add_column(
                    [beam.minor.to(u.arcsec)] * _nchan,
                    name="BMIN",
                )
                beams.add_column([beam.pa.to(u.deg)] * _nchan, name="BPA")

            except Exception as e:
                logger.error("Couldn't get beam from header")
                raise e

    nchan = len(beams)
    return beams, nchan, beamlog


def getfacs(
    beams: Beams, convbeams: Beams, dx: u.Quantity, dy: u.Quantity
) -> np.ndarray:
    """Get the conversion factors for each beam.

    Args:
        beams (Beams): Old beams.
        convbeams (Beams): Convolving beams.
        dx (u.Quantity): Pixel size in x direction.
        dy (u.Quantity): Pixel size in y direction.

    Returns:
        np.ndarray: Conversion factors.
    """
    facs_list = []
    for conbm, old_beam in zip(convbeams, beams):
        if conbm == Beam(major=0 * u.deg, minor=0 * u.deg, pa=0 * u.deg):
            fac = 1.0
        else:
            fac, amp, outbmaj, outbmin, outbpa = au2.gauss_factor(
                beamConv=[
                    conbm.major.to(u.arcsec).value,
                    conbm.minor.to(u.arcsec).value,
                    conbm.pa.to(u.deg).value,
                ],
                beamOrig=[
                    old_beam.major.to(u.arcsec).value,
                    old_beam.minor.to(u.arcsec).value,
                    old_beam.pa.to(u.deg).value,
                ],
                dx1=dx.to(u.arcsec).value,
                dy1=dy.to(u.arcsec).value,
            )
        facs_list.append(fac)
    facs = np.array(facs_list)
    return facs


def cpu_to_use(max_cpu: int, count: int) -> int:
    """Find number of cpus to use.
    Find the right number of cpus to use when dividing up a task, such
    that there are no remainders.
    Args:
        max_cpu (int): Maximum number of cores to use for a process.
        count (int): Number of tasks.

    Returns:
        int: Maximum number of cores to be used that divides into the number
        of tasks.
    """
    factors_list = []
    for i in range(1, count + 1):
        if count % i == 0:
            factors_list.append(i)
    factors = np.array(factors_list)
    return max(factors[factors <= max_cpu])


def worker(
    filename: str,
    idx: int,
    **kwargs,
) -> np.ndarray:
    """Smooth worker function.

    Extracts a single image from a FITS cube and smooths it.

    Args:
        filename (str): FITS cube filename.
        idx (int): Channel index.

    Kwargs:
        Passed to :func:`smooth`.

    Returns:
        np.ndarray: Smoothed channel image.
    """
    cube = SpectralCube.read(filename)
    # Get spectral axis
    wcs = cube.wcs
    axis_type_dict = wcs.get_axis_types()[::-1]  # Reverse order for fits
    axis_names = [i["coordinate_type"] for i in axis_type_dict]
    spec_idx = axis_names.index("spectral")
    slicer = [slice(None)] * len(cube.unmasked_data.shape)
    slicer[spec_idx] = idx
    slicer = tuple(slicer)

    plane = cube.unmasked_data[slicer].value.astype(np.float32)
    logger.debug(f"Size of plane is {(plane.nbytes*u.byte).to(u.MB)}")
    newim = smooth(image=plane, **kwargs)
    return newim


def makedata(files: List[str], outdir: str) -> Dict[str, dict]:
    """Create data dictionary.

    Args:
        files (List[str]): List of filenames.
        outdir (str): Output directory.

    Raises:
        Exception: If pixel grid in X and Y is not the same.

    Returns:
        Dict[Dict]: Data and metadata for each channel and image.
    """
    datadict = {}  # type: Dict[str,dict]
    for i, (file, out) in enumerate(zip(files, outdir)):
        # Set up files
        datadict[f"cube_{i}"] = {}
        datadict[f"cube_{i}"]["filename"] = file
        datadict[f"cube_{i}"]["outdir"] = out
        # Get metadata
        header = fits.getheader(file)
        w = WCS(header)
        pixelscales = proj_plane_pixel_scales(w)

        dxas = pixelscales[0] * u.deg
        dyas = pixelscales[1] * u.deg

        datadict[f"cube_{i}"]["dx"] = dxas
        datadict[f"cube_{i}"]["dy"] = dyas
        if not np.isclose(dxas, dyas):
            raise Exception("GRID MUST BE SAME IN X AND Y")
        # Get beam info
        beam, nchan, beamlog = getbeams(file=file, header=header)
        datadict[f"cube_{i}"]["beamlog"] = beamlog
        datadict[f"cube_{i}"]["beam"] = beam
        datadict[f"cube_{i}"]["nchan"] = nchan
    return datadict


def commonbeamer(
    datadict: Dict[str, dict],
    nchans: int,
    conv_mode: str = "robust",
    mode: str = "natural",
    suffix: str = None,
    target_beam: Beam = None,
    circularise: bool = False,
    tolerance: float = 0.0001,
    nsamps: int = 200,
    epsilon: float = 0.0005,
) -> Dict[str, dict]:
    """Find common beam for all channels.
    Computed beams will be written to convolving beam logger.

    Args:
        datadict (Dict[str, dict]): Main data dictionary.
        nchans (int): Number of channels.
        conv_mode (str, optional): Convolution mode. Defaults to "robust".
        mode (str, optional): Frequency mode. Defaults to "natural".
        target_beam (Beam, optional): Target PSF. Defaults to None.
        circularise (bool, optional): Circularise PSF. Defaults to False.
        tolerance (float, optional): Common beam tolerance. Defaults to 0.0001.
        nsamps (int, optional): Common beam samples. Defaults to 200.
        epsilon (float, optional): Common beam epsilon. Defaults to 0.0005.

    Raises:
        Exception: If convolving beam will be undersampled on pixel grid.

    Returns:
        Dict[str, dict]: Updated data dictionary.
    """
    if suffix is None:
        suffix = mode
    ### Natural mode ###
    if mode == "natural":
        big_beams = []
        for n in trange(
            nchans,
            desc="Constructing beams",
            disable=(logger.level > logging.INFO),
        ):
            majors_list = []
            minors_list = []
            pas_list = []
            for key in datadict.keys():
                major = datadict[key]["beams"][n].major
                minor = datadict[key]["beams"][n].minor
                pa = datadict[key]["beams"][n].pa
                if datadict[key]["mask"][n]:
                    major *= np.nan
                    minor *= np.nan
                    pa *= np.nan
                majors_list.append(major.to(u.arcsec).value)
                minors_list.append(minor.to(u.arcsec).value)
                pas_list.append(pa.to(u.deg).value)

            majors = np.array(majors_list)
            minors = np.array(minors_list)
            pas = np.array(pas_list)

            majors *= u.arcsec
            minors *= u.arcsec
            pas *= u.deg
            big_beams.append(Beams(major=majors, minor=minors, pa=pas))

        # Find common beams
        bmaj_common = []
        bmin_common = []
        bpa_common = []
        for beams in tqdm(
            big_beams,
            desc="Finding common beam per channel",
            disable=(logger.level > logging.INFO),
            total=nchans,
        ):
            if all(np.isnan(beams)):
                commonbeam = Beam(
                    major=np.nan * u.deg, minor=np.nan * u.deg, pa=np.nan * u.deg
                )
            else:
                try:
                    commonbeam = beams[~np.isnan(beams)].common_beam(
                        tolerance=tolerance,
                        nsamps=nsamps,
                        epsilon=epsilon,
                    )
                except BeamError:
                    logger.warn("Couldn't find common beam with defaults")
                    logger.warn("Trying again with smaller tolerance")

                    commonbeam = beams[~np.isnan(beams)].common_beam(
                        tolerance=tolerance * 0.1,
                        nsamps=nsamps,
                        epsilon=epsilon,
                    )
                # Round up values
                commonbeam = Beam(
                    major=my_ceil(commonbeam.major.to(u.arcsec).value, precision=1)
                    * u.arcsec,
                    minor=my_ceil(commonbeam.minor.to(u.arcsec).value, precision=1)
                    * u.arcsec,
                    pa=round_up(commonbeam.pa.to(u.deg), decimals=2),
                )

                grid = datadict[key]["dy"]
                if conv_mode != "robust":
                    # Get the minor axis of the convolving beams
                    minorcons = []
                    for beam in beams[~np.isnan(beams)]:
                        minorcons += [
                            commonbeam.deconvolve(beam).minor.to(u.arcsec).value
                        ]
                    minorcons = np.array(minorcons) * u.arcsec
                    samps = minorcons / grid.to(u.arcsec)
                    # Check that convolving beam will be Nyquist sampled
                    if any(samps.value < 2):
                        # Set the convolving beam to be Nyquist sampled
                        nyq_con_beam = Beam(
                            major=grid * 2, minor=grid * 2, pa=0 * u.deg
                        )
                        # Find new target based on common beam * Nyquist beam
                        # Not sure if this is best - but it works
                        nyq_beam = commonbeam.convolve(nyq_con_beam)
                        nyq_beam = Beam(
                            major=my_ceil(
                                nyq_beam.major.to(u.arcsec).value, precision=1
                            )
                            * u.arcsec,
                            minor=my_ceil(
                                nyq_beam.minor.to(u.arcsec).value, precision=1
                            )
                            * u.arcsec,
                            pa=round_up(nyq_beam.pa.to(u.deg), decimals=2),
                        )
                        logger.info(
                            f"Smallest common Nyquist sampled beam is: {nyq_beam!r}"
                        )

                        logger.warn("COMMON BEAM WILL BE UNDERSAMPLED!")
                        logger.warn("SETTING COMMON BEAM TO NYQUIST BEAM")
                        commonbeam = nyq_beam

            bmaj_common.append(commonbeam.major.to(u.arcsec).value)
            bmin_common.append(commonbeam.minor.to(u.arcsec).value)
            bpa_common.append(commonbeam.pa.to(u.deg).value)

        bmaj_common *= u.arcsec
        bmin_common *= u.arcsec
        bpa_common *= u.deg

        # Make Beams object
        commonbeams = Beams(major=bmaj_common, minor=bmin_common, pa=bpa_common)

    elif mode == "total":
        majors_list = []
        minors_list = []
        pas_list = []
        for key in datadict.keys():
            major = datadict[key]["beams"].major
            minor = datadict[key]["beams"].minor
            pa = datadict[key]["beams"].pa
            major[datadict[key]["mask"]] *= np.nan
            minor[datadict[key]["mask"]] *= np.nan
            pa[datadict[key]["mask"]] *= np.nan
            majors_list.append(major.to(u.arcsec).value)
            minors_list.append(minor.to(u.arcsec).value)
            pas_list.append(pa.to(u.deg).value)

        majors = np.array(majors_list).ravel()
        minors = np.array(minors_list).ravel()
        pas = np.array(pas_list).ravel()

        majors *= u.arcsec
        minors *= u.arcsec
        pas *= u.deg
        big_beams = Beams(major=majors, minor=minors, pa=pas)

        logger.info("Finding common beam across all channels")
        logger.info("This may take some time...")

        try:
            commonbeam = big_beams[~np.isnan(big_beams)].common_beam(
                tolerance=tolerance, nsamps=nsamps, epsilon=epsilon
            )
        except BeamError:
            logger.warn("Couldn't find common beam with defaults")
            logger.warn("Trying again with smaller tolerance")

            commonbeam = big_beams[~np.isnan(big_beams)].common_beam(
                tolerance=tolerance * 0.1, nsamps=nsamps, epsilon=epsilon
            )
        if target_beam is not None:
            commonbeam = target_beam
        else:
            # Round up values
            commonbeam = Beam(
                major=my_ceil(commonbeam.major.to(u.arcsec).value, precision=1)
                * u.arcsec,
                minor=my_ceil(commonbeam.minor.to(u.arcsec).value, precision=1)
                * u.arcsec,
                pa=round_up(commonbeam.pa.to(u.deg), decimals=2),
            )
        if conv_mode != "robust":
            # Get the minor axis of the convolving beams
            grid = datadict[key]["dy"]
            minorcons = []
            for beam in big_beams[~np.isnan(big_beams)]:
                minorcons += [commonbeam.deconvolve(beam).minor.to(u.arcsec).value]
            minorcons = np.array(minorcons) * u.arcsec
            samps = minorcons / grid.to(u.arcsec)
            # Check that convolving beam will be Nyquist sampled
            if any(samps.value < 2):
                # Set the convolving beam to be Nyquist sampled
                nyq_con_beam = Beam(major=grid * 2, minor=grid * 2, pa=0 * u.deg)
                # Find new target based on common beam * Nyquist beam
                # Not sure if this is best - but it works
                nyq_beam = commonbeam.convolve(nyq_con_beam)
                nyq_beam = Beam(
                    major=my_ceil(nyq_beam.major.to(u.arcsec).value, precision=1)
                    * u.arcsec,
                    minor=my_ceil(nyq_beam.minor.to(u.arcsec).value, precision=1)
                    * u.arcsec,
                    pa=round_up(nyq_beam.pa.to(u.deg), decimals=2),
                )
                logger.info(f"Smallest common Nyquist sampled beam is: {nyq_beam!r}")
                if target_beam is not None:
                    commonbeam = target_beam
                    if target_beam < nyq_beam:
                        logger.warn("TARGET BEAM WILL BE UNDERSAMPLED!")
                        raise Exception("CAN'T UNDERSAMPLE BEAM - EXITING")
                else:
                    logger.warn("COMMON BEAM WILL BE UNDERSAMPLED!")
                    logger.warn("SETTING COMMON BEAM TO NYQUIST BEAM")
                    commonbeam = nyq_beam

        # Make Beams object
        commonbeams = Beams(
            major=[commonbeam.major] * nchans * commonbeam.major.unit,
            minor=[commonbeam.minor] * nchans * commonbeam.minor.unit,
            pa=[commonbeam.pa] * nchans * commonbeam.pa.unit,
        )

    if circularise:
        logger.info("Circular beam requested, setting BMIN=BMAJ and BPA=0")
        commonbeams = Beams(
            major=commonbeams.major,
            minor=commonbeams.major,
            pa=commonbeams.pa * 0,
        )

    logger.info("Final beams are:")
    for i, commonbeam in enumerate(commonbeams):
        logger.info(f"Channel {i}: {commonbeam!r}")

    for key in tqdm(
        datadict.keys(),
        desc="Getting convolution data",
        disable=(logger.level > logging.INFO),
    ):
        # Get convolving beams
        conv_bmaj = []
        conv_bmin = []
        conv_bpa = []
        old_beams = datadict[key]["beams"]
        masks = datadict[key]["mask"]
        for commonbeam, old_beam, mask in zip(commonbeams, old_beams, masks):
            if mask:
                convbeam = Beam(
                    major=np.nan * u.deg, minor=np.nan * u.deg, pa=np.nan * u.deg
                )
            else:
                old_beam_check = Beam(
                    major=my_ceil(old_beam.major.to(u.arcsec).value, precision=1)
                    * u.arcsec,
                    minor=my_ceil(old_beam.minor.to(u.arcsec).value, precision=1)
                    * u.arcsec,
                    pa=round_up(old_beam.pa.to(u.deg), decimals=2),
                )
                if commonbeam == old_beam_check:
                    convbeam = Beam(
                        major=0 * u.deg,
                        minor=0 * u.deg,
                        pa=0 * u.deg,
                    )
                    logger.warn(
                        f"New beam {commonbeam!r} and old beam {old_beam_check!r} are the same. Won't attempt convolution."
                    )

                else:
                    convbeam = commonbeam.deconvolve(old_beam)

            conv_bmaj.append(convbeam.major.to(u.arcsec).value)
            conv_bmin.append(convbeam.minor.to(u.arcsec).value)
            conv_bpa.append(convbeam.pa.to(u.deg).value)

        conv_bmaj *= u.arcsec
        conv_bmin *= u.arcsec
        conv_bpa *= u.deg

        # Construct beams object
        convbeams = Beams(major=conv_bmaj, minor=conv_bmin, pa=conv_bpa)

        # Get gaussian beam factors
        facs = getfacs(
            beams=datadict[key]["beams"],
            convbeams=convbeams,
            dx=datadict[key]["dx"],
            dy=datadict[key]["dy"],
        )
        datadict[key]["facs"] = facs

        # Setup conv beamlog
        datadict[key]["convbeams"] = convbeams
        commonbeam_log = datadict[key]["beamlog"].replace(".txt", f".{suffix}.txt")
        datadict[key]["commonbeams"] = commonbeams
        datadict[key]["commonbeamlog"] = commonbeam_log

        commonbeam_tab = Table()
        # Save target
        commonbeam_tab.add_column(np.arange(nchans), name="Channel")
        commonbeam_tab.add_column(commonbeams.major, name="Target BMAJ")
        commonbeam_tab.add_column(commonbeams.minor, name="Target BMIN")
        commonbeam_tab.add_column(commonbeams.pa, name="Target BPA")
        # Save convolving beams
        commonbeam_tab.add_column(convbeams.major, name="Convolving BMAJ")
        commonbeam_tab.add_column(convbeams.minor, name="Convolving BMIN")
        commonbeam_tab.add_column(convbeams.pa, name="Convolving BPA")
        # Save facs
        commonbeam_tab.add_column(facs, name="Convolving factor")

        # Write to log file
        units = ""
        for col in commonbeam_tab.colnames:
            unit = commonbeam_tab[col].unit
            unit = str(unit)
            units += unit + " "
        commonbeam_tab.meta["comments"] = [units]
        ascii.write(
            commonbeam_tab,
            output=commonbeam_log,
            format="commented_header",
            overwrite=True,
        )
        logger.info(f"Convolving log written to {commonbeam_log}")

    return datadict


def masking(nchans: int, datadict: dict, cutoff: u.Quantity = None) -> dict:
    """Apply masking to data.

    Args:
        nchans (int): Number of channels in cubes.
        datadict (dict): Data dictionary.
        cutoff (None, optional): Cutoff BMAJ size for masking. Defaults to None.

    Returns:
        dict: Updated data dictionary.
    """
    for key in datadict.keys():
        mask = np.array([False] * nchans)
        datadict[key]["mask"] = mask
    if cutoff is not None:
        for key in datadict.keys():
            majors = datadict[key]["beams"].major
            cutmask = majors > cutoff
            datadict[key]["mask"] += cutmask

    # Check for pipeline masking
    nullbeam = Beam(major=0 * u.deg, minor=0 * u.deg, pa=0 * u.deg)
    tiny = np.finfo(np.float32).tiny  # Smallest positive number - used to mask
    smallbeam = Beam(major=tiny * u.deg, minor=tiny * u.deg, pa=tiny * u.deg)
    for key in datadict.keys():
        nullmask = np.logical_or(
            datadict[key]["beams"] == nullbeam,
            datadict[key]["beams"] == smallbeam,
        )
        datadict[key]["mask"] += nullmask
    return datadict


def initfiles(
    filename: str,
    commonbeams: Beams,
    outdir: str,
    mode: str,
    suffix=None,
    prefix=None,
    ref_chan=None,
):
    """Initialise output files

    Args:
        datadict (dict): Main data dict - indexed
        mode (str): 'total' or 'natural'

    Returns:
        datadict: Updated datadict
    """
    logger.debug(f"Reading {filename}")
    with fits.open(filename, memmap=True, mode="denywrite") as hdulist:
        primary_hdu = hdulist[0]
        data = primary_hdu.data
        header = primary_hdu.header
        wcs = WCS(header)

    ## Header
    spec_axis = wcs.spectral
    crpix = int(spec_axis.wcs.crpix)
    nchans = spec_axis.array_shape[0]
    assert nchans == len(
        commonbeams
    ), "Number of channels in header and commonbeams do not match"
    chans = np.arange(nchans)
    if ref_chan is None:
        # Get reference channel, and attach PSF there
        crindex = crpix - 1  # For python!
    elif ref_chan == "first":
        crindex = 0
    elif ref_chan == "last":
        crindex = -1
    elif ref_chan == "mid":
        # Locate mid Channel
        # In python's 0-based index, the following will set the mid Channel to
        # the upper-mid value for even number of channels. For odd-number of
        # channels, the mid value is unique.
        crindex = nchans // 2
    ref_psf = commonbeams[crindex]
    # Check the Stokes
    stokes_axis = wcs.sub(["stokes"])
    if stokes_axis.array_shape == ():
        raise ValueError("No Stokes axis found")
    nstokes = stokes_axis.array_shape[0]
    if nstokes > 1:
        logger.critical(
            f"More than one Stokes parameter in header. Only the first one will be used."
        )
    pols = np.zeros_like(chans)  # Zeros because we take the first one
    if any(
        (
            np.isnan(ref_psf.major.value),
            np.isnan(ref_psf.minor.value),
            np.isnan(ref_psf.pa.value),
        )
    ):
        logger.warning("Reference PSF is NaN - replacing with 0 in the header")
        ref_psf = Beam(major=0 * u.deg, minor=0 * u.deg, pa=0 * u.deg)
        header["COMMENT"] = "Reference PSF is NaN"
        header["COMMENT"] = "- This is likely because the reference channel is masked."
        header["COMMENT"] = "- It has been replaced with 0 to keep FITS happy."
    header = ref_psf.attach_to_header(header)
    primary_hdu = fits.PrimaryHDU(data=data, header=header)
    if mode == "natural":
        # Make a CASA beamtable
        header["CASAMBM"] = True
        header["COMMENT"] = "The PSF in each image plane varies."
        header[
            "COMMENT"
        ] = "Full beam information is stored in the second FITS extension."
        tiny = np.finfo(np.float32).tiny
        beam_table = Table(
            data=[
                # Replace NaNs with np.finfo(np.float32).tiny - this is the smallest
                # positive number that can be represented in float32
                # We use this to keep CASA happy
                np.nan_to_num(commonbeams.major.to(u.arcsec), nan=tiny * u.arcsec),
                np.nan_to_num(commonbeams.minor.to(u.arcsec), nan=tiny * u.arcsec),
                np.nan_to_num(commonbeams.pa.to(u.deg), nan=tiny * u.deg),
                chans,
                pols,
            ],
            names=["BMAJ", "BMIN", "BPA", "CHAN", "POL"],
            dtype=["f4", "f4", "f4", "i4", "i4"],
        )
        header["COMMENT"] = f"The value '{tiny}' repsenents a NaN PSF in the beamtable."
        primary_hdu = fits.PrimaryHDU(data=data, header=header)
        tab_hdu = fits.table_to_hdu(beam_table)
        tab_header = tab_hdu.header
        tab_header["EXTNAME"] = "BEAMS"
        tab_header["NCHAN"] = nchans
        tab_header["NPOL"] = 1  # Only one pol for now
        new_hdulist = fits.HDUList([primary_hdu, tab_hdu])

    elif mode == "total":
        new_hdulist = fits.HDUList([primary_hdu])

    # Set up output file
    if suffix is None:
        suffix = mode
    outname = os.path.basename(filename)
    outname = outname.replace(".fits", "") + f".{suffix}.fits"
    if prefix is not None:
        outname = prefix + outname

    outfile = os.path.join(outdir, outname)
    logger.info(f"Initialising to {outfile}")
    new_hdulist.writeto(outfile, overwrite=True)

    return outfile


def readlogs(commonbeam_log: str) -> Tuple[Beams, Beams, np.ndarray]:
    """Read convolving log files

    Args:
        commonbeam_log (str): Filename of the common beam log

    Raises:
        Exception: If the log file is not found

    Returns:
        Tuple[Beams, Beams, np.ndarray]: Common beams, convolving beams, and scaling factors
    """
    logger.info(f"Reading from {commonbeam_log}")
    try:
        commonbeam_tab = Table.read(commonbeam_log, format="ascii.commented_header")
    except FileNotFoundError:
        raise Exception("beamlogConvolve must be co-located with image")
    # Convert to Beams
    commonbeams = Beams(
        major=commonbeam_tab["Target BMAJ"] * u.arcsec,
        minor=commonbeam_tab["Target BMIN"] * u.arcsec,
        pa=commonbeam_tab["Target BPA"] * u.deg,
    )
    convbeams = Beams(
        major=commonbeam_tab["Convolving BMAJ"] * u.arcsec,
        minor=commonbeam_tab["Convolving BMIN"] * u.arcsec,
        pa=commonbeam_tab["Convolving BPA"] * u.deg,
    )
    facs = np.array(commonbeam_tab["Convolving factor"])
    logger.info("Final beams are:")
    for i, commonbeam in enumerate(commonbeams):
        logger.info(f"Channel {i}: {commonbeam!r}")
    return commonbeams, convbeams, facs


def main(
    infile: list,
    uselogs: bool = False,
    mode: str = "natural",
    conv_mode: str = "robust",
    dryrun: bool = False,
    prefix: str = None,
    suffix: str = None,
    outdir: str = None,
    bmaj: float = None,
    bmin: float = None,
    bpa: float = None,
    cutoff: float = None,
    circularise: bool = False,
    ref_chan: int = None,
    tolerance: float = 0.0001,
    epsilon: float = 0.0005,
    nsamps: int = 200,
):
    """main script

    Args:
        args (args): Command line args

    """

    if myPE == 0:
        logger.info(f"Total number of MPI ranks = {nPE}")
        # Parse args
        if dryrun:
            logger.info("Doing a dry run -- no files will be saved")

        # Check mode
        logger.info(f"Mode is {mode}")
        if mode == "natural" and mode == "total":
            raise Exception("'mode' must be 'natural' or 'total'")
        if mode == "natural":
            logger.info("Smoothing each channel to a common resolution")
        if mode == "total":
            logger.info("Smoothing all channels to a common resolution")

        # Check cutoff
        if cutoff is not None:
            cutoff *= u.arcsec
            logger.info(f"Cutoff is: {cutoff}")

        # Check target
        logger.debug(conv_mode)
        if (
            not conv_mode == "robust"
            and not conv_mode == "scipy"
            and not conv_mode == "astropy"
            and not conv_mode == "astropy_fft"
        ):
            raise Exception("Please select valid convolution method!")

        logger.info(f"Using convolution method {conv_mode}")
        if conv_mode == "robust":
            logger.info("This is the most robust method. And fast!")
        elif conv_mode == "scipy":
            logger.info("This fast, but not robust to NaNs or small PSF changes")
        else:
            logger.info(
                "This is slower, but robust to NaNs, but not to small PSF changes"
            )

        nonetest = [test is None for test in [bmaj, bmin, bpa]]

        if not all(nonetest) and mode != "total":
            raise Exception("Only specify a target beam in 'total' mode")

        if all(nonetest):
            target_beam = None

        elif not all(nonetest) and any(nonetest):
            raise Exception("Please specify all target beam params!")

        elif not all(nonetest) and not any(nonetest):
            target_beam = Beam(bmaj * u.arcsec, bmin * u.arcsec, bpa * u.deg)
            logger.info(f"Target beam is {target_beam!r}")

        files = sorted(infile)
        if files == []:
            raise Exception("No files found!")

        if outdir is not None:
            if outdir[-1] == "/":
                outdir = outdir[:-1]
            outdir = [outdir] * len(files)
        else:
            outdir = []
            for f in files:
                out = os.path.dirname(f)
                if out == "":
                    out = "."
                outdir += [out]

        datadict = makedata(files, outdir)

        # Sanity check channel counts
        nchans = np.array([datadict[key]["nchan"] for key in datadict.keys()])
        check = all(nchans == nchans[0])

        if not check:
            raise Exception("Unequal number of spectral channels!")

        else:
            nchans = nchans[0]

        # Check suffix
        if suffix is None:
            suffix = mode

        # Construct Beams objects
        for key in datadict.keys():
            beam = datadict[key]["beam"]
            bmaj = np.array(beam["BMAJ"]) * beam["BMAJ"].unit
            bmin = np.array(beam["BMIN"]) * beam["BMIN"].unit
            bpa = np.array(beam["BPA"]) * beam["BPA"].unit
            beams = Beams(major=bmaj, minor=bmin, pa=bpa)
            datadict[key]["beams"] = beams

        # Apply some masking
        datadict = masking(nchans=nchans, datadict=datadict, cutoff=cutoff)

        if not uselogs:
            datadict = commonbeamer(
                datadict=datadict,
                nchans=nchans,
                conv_mode=conv_mode,
                target_beam=target_beam,
                mode=mode,
                suffix=suffix,
                circularise=circularise,
                tolerance=tolerance,
                nsamps=nsamps,
                epsilon=epsilon,
            )
        else:
            logger.info("Reading from convolve beamlog files")
            for key in datadict.keys():
                commonbeam_log = datadict[key]["beamlog"].replace(
                    ".txt", f".{suffix}.txt"
                )
                commonbeams, convbeams, facs = readlogs(commonbeam_log)
                # Save to datadict
                datadict[key]["facs"] = facs
                datadict[key]["convbeams"] = convbeams
                datadict[key]["commonbeams"] = commonbeams
                datadict[key]["commonbeamlog"] = commonbeam_log
    else:
        files = None
        datadict = None
        nchans = None

    if dryrun:
        logger.info("Doing a dryrun so all done!")
        return datadict

    # Init the files in parallel
    if myPE == 0:
        logger.info("Initialising output files")
    if mpiSwitch:
        files = comm.bcast(files, root=0)
        datadict = comm.bcast(datadict, root=0)
        nchans = comm.bcast(nchans, root=0)

    inputs = list(datadict.keys())
    dims = len(inputs)

    if nPE > dims:
        my_start = myPE
        my_end = myPE

    else:
        count = dims // nPE
        rem = dims % nPE
        if myPE < rem:
            # The first 'remainder' ranks get 'count + 1' tasks each
            my_start = myPE * (count + 1)
            my_end = my_start + count

        else:
            # The remaining 'size - remainder' ranks get 'count' task each
            my_start = myPE * count + rem
            my_end = my_start + (count - 1)

    if myPE == 0:
        logger.info(f"There are {dims} files to init")
    logger.debug(f"My start is {my_start}, my end is {my_end}")

    # Init output files and retrieve file names
    outfile_dict = {}
    for inp in inputs[my_start : my_end + 1]:
        outfile = initfiles(
            filename=datadict[inp]["filename"],
            commonbeams=datadict[inp]["commonbeams"],
            outdir=datadict[inp]["outdir"],
            mode=mode,
            suffix=suffix,
            prefix=prefix,
            ref_chan=ref_chan,
        )
        outfile_dict.update({inp: outfile})

    if mpiSwitch:
        # Send to master proc
        outlist = comm.gather(outfile_dict, root=0)
    else:
        outlist = [outfile_dict]

    if mpiSwitch:
        comm.Barrier()

    # Now do the convolution in parallel
    if myPE == 0:
        # Conver list to dict and save to main dict
        outlist_dict = {}
        for d in outlist:
            outlist_dict.update(d)
        # Also make inputs list
        inputs = []
        for key in datadict.keys():
            datadict[key]["outfile"] = outlist_dict[key]
            for chan in range(nchans):
                inputs.append((key, chan))

    else:
        datadict = None
        inputs = None
    if mpiSwitch:
        comm.Barrier()
    if mpiSwitch:
        inputs = comm.bcast(inputs, root=0)
        datadict = comm.bcast(datadict, root=0)

    dims = len(files) * nchans
    assert len(inputs) == dims
    count = dims // nPE
    rem = dims % nPE
    if myPE < rem:
        # The first 'remainder' ranks get 'count + 1' tasks each
        my_start = myPE * (count + 1)
        my_end = my_start + count

    else:
        # The remaining 'size - remainder' ranks get 'count' task each
        my_start = myPE * count + rem
        my_end = my_start + (count - 1)
    if myPE == 0:
        logger.info(f"There are {nchans} channels, across {len(files)} files")
    logger.debug(f"My start is {my_start}, my end is {my_end}")

    for inp in inputs[my_start : my_end + 1]:
        key, chan = inp
        outfile = datadict[key]["outfile"]
        logger.debug(f"{outfile}  - channel {chan} - Started")

        cubedict = datadict[key]
        newim = worker(
            filename=cubedict["filename"],
            idx=chan,
            dx=cubedict["dx"],
            dy=cubedict["dy"],
            old_beam=cubedict["beams"][chan],
            final_beam=cubedict["commonbeams"][chan],
            conbeam=cubedict["convbeams"][chan],
            sfactor=cubedict["facs"][chan],
            conv_mode=conv_mode,
        )

        with fits.open(outfile, mode="update", memmap=True) as outfh:
            # Find which axis is the spectral and stokes
            wcs = WCS(outfh[0])
            axis_type_dict = wcs.get_axis_types()[::-1]  # Reverse order for fits
            axis_names = [i["coordinate_type"] for i in axis_type_dict]
            spec_idx = axis_names.index("spectral")
            stokes_idx = axis_names.index("stokes")
            slicer = [slice(None)] * len(outfh[0].data.shape)
            slicer[spec_idx] = chan
            slicer[stokes_idx] = 0  # only do single stokes
            slicer = tuple(slicer)

            outfh[0].data[slicer] = newim.astype(np.float32)  # make sure data is 32-bit
            outfh.flush()
        logger.info(f"{outfile}  - channel {chan} - Done")

    logger.info("Done!")
    return datadict


def cli():
    """Command-line interface"""
    import argparse

    # Help string to be shown using the -h option
    descStr = """
    Smooth a field of 3D cubes to a common resolution.

    - Parallelisation is done using MPI.

    - Default names of output files are /path/to/beamlog{infile//.fits/.{SUFFIX}.fits}

    - By default, the smallest common beam will be automatically computed.
    - Optionally, you can specify a target beam to use.

    - It is currently assumed that cubes will be 4D with a dummy Stokes axis.
    - Iterating over Stokes axis is not yet supported.

    """

    # Parse the command line options
    parser = argparse.ArgumentParser(
        description=descStr, formatter_class=argparse.RawTextHelpFormatter
    )

    parser.add_argument(
        "infile",
        metavar="infile",
        type=str,
        help="""Input FITS image(s) to smooth (can be a wildcard)
        - CASA beamtable will be used if present i.e. if CASAMBM = T
        - Otherwise beam info must be in co-located beamlog files.
        - beamlog must have the name /path/to/beamlog{infile//.fits/.txt}
        """,
        nargs="+",
    )

    parser.add_argument(
        "--uselogs",
        dest="uselogs",
        action="store_true",
        help="Get convolving information from previous run [False].",
    )

    parser.add_argument(
        "--mode",
        dest="mode",
        type=str,
        default="natural",
        help="""Common resolution mode [natural].
        natural -- allow frequency variation.
        total -- smooth all plans to a common resolution.
        """,
    )

    parser.add_argument(
        "--conv_mode",
        dest="conv_mode",
        type=str,
        default="robust",
        choices=["robust", "scipy", "astropy", "astropy_fft"],
        help="""Which method to use for convolution [robust].
        'robust' computes the analytic FT of the convolving Gaussian.
        Note this mode can now handle NaNs in the data.
        Can also be 'scipy', 'astropy', or 'astropy_fft'.
        Note these other methods cannot cope well with small convolving beams.
        """,
    )

    parser.add_argument(
        "-v", "--verbosity", default=0, action="count", help="Increase output verbosity"
    )

    parser.add_argument(
        "--logfile",
        default=None,
        type=str,
        help="Save logging output to file",
    )

    parser.add_argument(
        "-d",
        "--dryrun",
        dest="dryrun",
        action="store_true",
        help="Compute common beam and stop [False].",
    )

    parser.add_argument(
        "-p",
        "--prefix",
        dest="prefix",
        type=str,
        default=None,
        help="Add prefix to output filenames.",
    )

    parser.add_argument(
        "-s",
        "--suffix",
        dest="suffix",
        type=str,
        default=None,
        help="Add suffix to output filenames [{MODE}].",
    )

    parser.add_argument(
        "-o",
        "--outdir",
        dest="outdir",
        type=str,
        default=None,
        help="Output directory of smoothed FITS image(s) [None - same as input].",
    )

    parser.add_argument(
        "--bmaj",
        dest="bmaj",
        type=float,
        default=None,
        help="BMAJ to convolve to [max BMAJ from given image(s)].",
    )

    parser.add_argument(
        "--bmin",
        dest="bmin",
        type=float,
        default=None,
        help="BMIN to convolve to [max BMAJ from given image(s)].",
    )

    parser.add_argument(
        "--bpa", dest="bpa", type=float, default=None, help="BPA to convolve to [0]."
    )

    parser.add_argument(
        "-c",
        "--cutoff",
        dest="cutoff",
        type=float,
        default=None,
        help="Cutoff BMAJ value (arcsec) -- Blank channels with BMAJ larger than this [None -- no limit]",
    )

    parser.add_argument(
        "--circularise",
        action="store_true",
        help="Circularise the final PSF -- Sets the BMIN = BMAJ, and BPA=0.",
    )

    parser.add_argument(
        "--ref_chan",
        dest="ref_chan",
        type=str,
        default=None,
        choices=["first", "last", "mid"],
        help="""Reference psf for header [None].
            first  -- use psf for first frequency channel.
            last -- use psf for the last frequency channel.
            mid -- use psf for the centre frequency channel.
            Will use the CRPIX channel if not set.
            """,
    )

    parser.add_argument(
        "-t",
        "--tolerance",
        dest="tolerance",
        type=float,
        default=0.0001,
        help="tolerance for radio_beam.commonbeam.",
    )

    parser.add_argument(
        "-e",
        "--epsilon",
        dest="epsilon",
        type=float,
        default=0.0005,
        help="epsilon for radio_beam.commonbeam.",
    )

    parser.add_argument(
        "-n",
        "--nsamps",
        dest="nsamps",
        type=int,
        default=200,
        help="nsamps for radio_beam.commonbeam.",
    )

    args = parser.parse_args()

    # Set up logging
    logger = setup_logger(
        verbosity=args.verbosity,
        filename=args.logfile,
    )

    arg_dict = vars(args)
    # Pop the verbosity argument
    _ = arg_dict.pop("verbosity")
    # Pop the logfile argument
    _ = arg_dict.pop("logfile")

    _ = main(**arg_dict)


if __name__ == "__main__":
    sys.exit(cli())



#!/usr/bin/env python3
"""
Code to generate FT of final 2D-Gaussian to be used
for convolving an image. The code deconvolves the input
psf. The intrinsic psf must be specified.

Python version of gaussft.f by Wasim Raja
"""
from typing import Tuple

import numba as nb
import numpy as np


@nb.njit(cache=True)
def gaussft(
    bmin_in: float,
    bmaj_in: float,
    bpa_in: float,
    bmin: float,
    bmaj: float,
    bpa: float,
    u: np.ndarray,
    v: np.ndarray,
) -> Tuple[np.ndarray, float]:
    """
    Compute the Fourier transform of a 2D Gaussian for convolution.

    Parameters:
        bmin_in (float): Intrinsic psf BMIN (degrees)
        bmaj_in (float): Intrinsic psf BMAJ (degrees)
        bpa_in (float): Intrinsic psf BPA (degrees)
        bmin (float): Final psf BMIN (degrees)
        bmaj (float): Final psf BMAJ (degrees)
        bpa (float): Final psf BPA (degrees)
        u (np.ndarray): Fourier coordinates corresponding to image coord x
        v (np.ndarray): Fourier coordinates corresponding to image coord y

    Returns:
        g_final (np.ndarray): Final array to be multiplied to FT(image) for convolution
            in the FT domain.
        g_ratio (float): Factor for flux scaling
    """
    deg2rad = np.pi / 180.0

    bmaj_in_rad, bmin_in_rad, bpa_in_rad = (
        bmaj_in * deg2rad,
        bmin_in * deg2rad,
        bpa_in * deg2rad,
    )
    bmaj_rad, bmin_rad, bpa_rad = bmaj * deg2rad, bmin * deg2rad, bpa * deg2rad

    sx, sy = bmaj_rad / (2 * np.sqrt(2.0 * np.log(2.0))), bmin_rad / (
        2 * np.sqrt(2.0 * np.log(2.0))
    )
    sx_in, sy_in = bmaj_in_rad / (2.0 * np.sqrt(2.0 * np.log(2.0))), bmin_in_rad / (
        2.0 * np.sqrt(2.0 * np.log(2.0))
    )

    u_cosbpa, u_sinbpa = u * np.cos(bpa_rad), u * np.sin(bpa_rad)
    u_cosbpa_in, u_sinbpa_in = u * np.cos(bpa_in_rad), u * np.sin(bpa_in_rad)

    v_cosbpa, v_sinbpa = v * np.cos(bpa_rad), v * np.sin(bpa_rad)
    v_cosbpa_in, v_sinbpa_in = v * np.cos(bpa_in_rad), v * np.sin(bpa_in_rad)

    g_amp = np.sqrt(2.0 * np.pi * sx * sy)

    dg_amp = np.sqrt(2.0 * np.pi * sx_in * sy_in)

    g_ratio = g_amp / dg_amp

    # Vectorized calculation of ur, vr, g_arg, and dg_arg
    ur = u_cosbpa[:, np.newaxis] - v_sinbpa[np.newaxis, :]
    vr = u_sinbpa[:, np.newaxis] + v_cosbpa[np.newaxis, :]
    g_arg = -2.0 * np.pi**2 * ((sx * ur) ** 2 + (sy * vr) ** 2)

    ur_in = u_cosbpa_in[:, np.newaxis] - v_sinbpa_in[np.newaxis, :]
    vr_in = u_sinbpa_in[:, np.newaxis] + v_cosbpa_in[np.newaxis, :]
    dg_arg = -2.0 * np.pi**2 * ((sx_in * ur_in) ** 2 + (sy_in * vr_in) ** 2)

    # Vectorized calculation of g_final
    g_final = g_ratio * np.exp(g_arg - dg_arg)

    return g_final, g_ratio



#!/usr/bin/env python3
""" Find bad channels by checking statistics of each channel image. """

import argparse
import logging
import warnings
from typing import List, Tuple, Union

import astropy.units as u
import numpy as np
from astropy.stats import mad_std
from spectral_cube import SpectralCube
from spectral_cube.utils import SpectralCubeWarning

from racs_tools.logging import logger, setup_logger

warnings.filterwarnings(action="ignore", category=SpectralCubeWarning, append=True)


#############################################
####### ADAPTED FROM SCRIPT BY G. HEALD #####
#############################################
def getcube(filename: str) -> SpectralCube:
    """Read FITS file as SpectralCube
    Masks out 0Jy/beam pixels

    Args:
        filename (str): Filename

    Returns:
        SpectralCube: Data cube
    """
    cube = SpectralCube.read(filename)
    mask = ~(cube == 0 * u.jansky / u.beam)
    cube = cube.with_mask(mask)
    return cube


def getbadchans(
    qcube: SpectralCube,
    ucube: SpectralCube,
    cliplev: float = 5,
) -> Tuple[np.ndarray, u.Quantity, u.Quantity]:
    """Find bad channels in Stokes Q and U cubes

    Args:
        qcube (SpectralCube): Stokes Q data
        ucube (SpectralCube): Stokes U data
        cliplev (float, optional): Number stddev above median to clip. Defaults to 5.

    Returns:
        Tuple[np.ndarray, u.Quantity, u.Quantity]: Bad channels in Q, U, and noise spectra
    """
    assert len(ucube.spectral_axis) == len(qcube.spectral_axis)

    qnoisevals = (
        qcube.apply_function_parallel_spatial(
            function=mad_std,
            parallel=False,
            use_memmap=True,
            ignore_nan=True,
        )[:, 0, 0].unitless_filled_data[:]
        * qcube.unit
    )
    unoisevals = (
        ucube.apply_function_parallel_spatial(
            function=mad_std,
            parallel=False,
            use_memmap=True,
            ignore_nan=True,
        )[:, 0, 0].unitless_filled_data[:]
        * ucube.unit
    )
    qmeannoise = np.nanmedian(qnoisevals)
    qstdnoise = mad_std(qnoisevals, ignore_nan=True)
    logger.info(
        f"Median Q noise=({qmeannoise.value:0.3f}±{qstdnoise.value:0.3f}) / ({qmeannoise.unit})"
    )
    umeannoise = np.nanmedian(unoisevals)
    ustdnoise = mad_std(unoisevals, ignore_nan=True)
    logger.info(
        f"Median U noise=({umeannoise.value:0.3f}±{ustdnoise.value:0.3f}) / ({umeannoise.unit})"
    )
    qbadones = np.logical_or(
        qnoisevals > (qmeannoise + cliplev * qstdnoise), ~np.isfinite(qnoisevals)
    )
    ubadones = np.logical_or(
        unoisevals > (umeannoise + cliplev * ustdnoise), ~np.isfinite(unoisevals)
    )
    logger.info(f"{qbadones.sum()} of {len(qcube.spectral_axis)} are bad in Q")
    logger.info(f"{ubadones.sum()} of {len(ucube.spectral_axis)} are bad in U")
    total_bad = np.logical_or(qbadones, ubadones)
    logger.info(f"{total_bad.sum()} of {len(qcube.spectral_axis)} are bad in Q -or- U")

    return total_bad, qnoisevals, unoisevals


def blankchans(
    qcube: SpectralCube, ucube: SpectralCube, totalbad: np.ndarray, blank: bool = False
) -> Tuple[SpectralCube, SpectralCube]:
    """Mask out bad channels

    Args:
        qcube (SpectralCube): Stokes Q data
        ucube (SpectralCube): Stokes U data
        totalbad (np.ndarray): Flagged bad channels
        blank (bool, optional): Print flags to screen. Defaults to False.

    Returns:
        Tuple[SpectralCube, SpectralCube]: _description_
    """
    chans = np.arange(len(qcube.spectral_axis))
    badchans = chans[totalbad]
    badfreqs = qcube.spectral_axis[totalbad]
    if not blank:
        logger.info(
            "Nothing will be blanked, but these are the channels/frequencies that would be with the -b option activated:"
        )
    logger.info(f"Bad channels are {badchans}")
    logger.info(f"Bad frequencies are {badfreqs}")
    totalgood = ~totalbad
    q_msk = qcube.mask_channels(totalgood)
    u_msk = ucube.mask_channels(totalgood)
    return q_msk, u_msk


def writefits(qcube: SpectralCube, ucube: SpectralCube, qfile: str, ufile: str) -> None:
    """Write output to disk

    Args:
        qcube (SpectralCube): Stokes Q data
        ucube (SpectralCube): Stokes U data
        qfile (str): Original Q file
        ufile (str): Original U file
    """
    outfile = qfile.replace(".fits", ".blanked.fits")
    logger.info(f"Writing to {outfile}")
    qcube.write(outfile, format="fits", overwrite=True)
    outfile = ufile.replace(".fits", ".blanked.fits")
    logger.info(f"Writing to {outfile}")
    ucube.write(outfile, format="fits", overwrite=True)


def main(
    qfile: str,
    ufile: str,
    blank: bool = False,
    cliplev: float = 5,
    iterate: int = 1,
    outfile: Union[str, None] = None,
    save_noise: bool = False,
) -> None:
    """Flag bad channels in Stokes Q and U cubes

    Args:
        qfile (str): Stokes Q fits file
        ufile (str): Stokes U fits file
        blank (bool, optional): Flag bad data and save to disk. Defaults to False.
        cliplev (float, optional): Number of stddev above median to flag. Defaults to 5.
        iterate (int, optional): Number of flagging iterations. Defaults to 1.
        outfile (str, optional): File to write flagged channels to. Defaults to None.
    """
    qcube = getcube(qfile)
    ucube = getcube(ufile)

    assert len(ucube.spectral_axis) == len(
        qcube.spectral_axis
    ), "Cubes have different number of channels"

    # Iterate
    for i in range(iterate):
        logger.info(f"Flagging iteration {i+1} of {iterate}")
        totalbad, qnoisevals, unoisevals = getbadchans(
            qcube,
            ucube,
            cliplev=cliplev,
        )
        qcube, ucube = blankchans(
            qcube=qcube, ucube=ucube, totalbad=totalbad, blank=blank
        )

    if save_noise:
        logger.info("Saving noise values to disk")
        np.savetxt(qfile.replace(".fits", ".noise.txt"), qnoisevals.value)
        np.savetxt(ufile.replace(".fits", ".noise.txt"), unoisevals.value)

    if blank:
        logger.warning("Overwriting original files")
        writefits(qcube=qcube, ucube=ucube, qfile=qfile, ufile=ufile)

    if outfile:
        logger.info(f"Saving bad files to {outfile}")
        np.savetxt(outfile, totalbad)


def cli() -> None:
    warnings.filterwarnings(
        "ignore",
        message="Cube is a Stokes cube, returning spectral cube for I component",
    )
    warnings.filterwarnings("ignore", message="Invalid value encountered in less")
    warnings.filterwarnings("ignore", message="Invalid value encountered in greater")
    warnings.filterwarnings("ignore", message="overflow encountered in square")
    warnings.filterwarnings("ignore", message="All-NaN slice encountered")

    # Parse the command line options
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument("qfile", type=str, help="Stokes Q fits file")
    parser.add_argument("ufile", type=str, help="Stokes U fits file")
    parser.add_argument(
        "-s",
        "--save_noise",
        help="Save noise values to disk [default False]",
        action="store_true",
    )
    parser.add_argument(
        "-b",
        "--blank",
        help="Blank bad channels? [default False - just print out bad channels]",
        action="store_true",
    )
    parser.add_argument(
        "-c",
        "--cliplev",
        help="Clip level in sigma, make this number lower to be more aggressive [default 5]",
        default=5.0,
        type=float,
    )

    parser.add_argument(
        "-i",
        "--iterate",
        help="Iterate flagging check N times [dafult 1 -- one pass only]",
        default=1,
        type=int,
    )

    parser.add_argument(
        "-f",
        "--file",
        help="Filename to write bad channel indices to file [None --  do not write]",
        default=None,
        type=str,
    )
    parser.add_argument(
        "-v", "--verbosity", default=0, action="count", help="Increase output verbosity"
    )
    parser.add_argument(
        "--logfile",
        default=None,
        type=str,
        help="Save logging output to file",
    )

    args = parser.parse_args()

    # Set up logging
    logger = setup_logger(
        verbosity=args.verbosity,
        filename=args.logfile,
    )

    main(
        qfile=args.qfile,
        ufile=args.ufile,
        blank=args.blank,
        cliplev=args.cliplev,
        iterate=args.iterate,
        outfile=args.file,
        save_noise=args.save_noise,
    )


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
""" Fast convolution in the UV domain """
__author__ = "Wasim Raja"


from typing import Tuple

import astropy.units as units
import numpy as np
import scipy.signal
from astropy import convolution
from astropy import units as u
from radio_beam import Beam

import racs_tools.gaussft as gaussft
from racs_tools.logging import logger


def convolve(
    image: np.ndarray, old_beam: Beam, new_beam: Beam, dx: u.Quantity, dy: u.Quantity
) -> Tuple[np.ndarray, float]:
    """Convolve by X-ing in the Fourier domain.
        - convolution with Gaussian kernels only
        - no need for generation of a kernel image
        - direct computation of the FT of the kernel
        - dimension of FT = dimension of image

    Args:
        image (np.ndarray): The image to be convolved.
        old_beam (Beam): Current image PSF.
        new_beam (Beam): Target image PSF.
        dx (u.Quantity): Grid size in x (e.g. CDELT1)
        dy (u.Quantity): Grid size in y (e.g. CDELT2)

    Returns:
        Tuple[np.ndarray, float]: (convolved image, scaling factor)
    """

    nanflag = np.isnan(image).any()

    if nanflag:
        # Create a mask for the NaNs
        mask = np.isnan(image).astype(int)
        logger.warning(
            f"Image contains {mask.sum()} ({mask.sum()/ mask.size *100 :0.1f}%) NaNs"
        )
        image = np.nan_to_num(image)

    nx = image.shape[0]
    ny = image.shape[1]

    # The coordinates in FT domain:
    u_image = np.fft.fftfreq(nx, d=dx.to(units.rad).value)
    v_image = np.fft.fftfreq(ny, d=dy.to(units.rad).value)

    g_final = np.zeros((nx, ny), dtype=float)
    [g_final, g_ratio] = gaussft.gaussft(
        bmin_in=old_beam.minor.to(units.deg).value,
        bmaj_in=old_beam.major.to(units.deg).value,
        bpa_in=old_beam.pa.to(units.deg).value,
        bmin=new_beam.minor.to(units.deg).value,
        bmaj=new_beam.major.to(units.deg).value,
        bpa=new_beam.pa.to(units.deg).value,
        u=u_image,
        v=v_image,
    )
    # Perform the x-ing in the FT domain
    im_f = np.fft.fft2(image)

    # Now convolve with the desired Gaussian:
    M = np.multiply(im_f, g_final)
    im_conv = np.fft.ifft2(M)
    im_conv = np.real(im_conv)

    if nanflag:
        # Convert the mask to the FT domain
        mask_f = np.fft.fft2(mask)
        # Multiply the mask by the FT of the Gaussian
        M = np.multiply(mask_f, g_final)
        # Invert the FT of the mask
        mask_conv = np.fft.ifft2(M)
        mask_conv = np.real(mask_conv)
        # Use approx values to find the NaNs
        # Need this to get around numerical issues
        mask_conv = ~(mask_conv + 1 < 2)
        logger.warning(
            f"Convolved image contains {mask_conv.sum()} ({mask_conv.sum()/ mask_conv.size *100 :0.1f}%) NaNs"
        )
        im_conv[mask_conv > 0] = np.nan

    return im_conv, g_ratio


def smooth(
    image: np.ndarray,
    old_beam: Beam,
    final_beam: Beam,
    dx: u.Quantity,
    dy: u.Quantity,
    sfactor: float,
    conbeam: Beam = None,
    conv_mode: str = "robust",
) -> np.ndarray:
    """Apply smoothing to image

    Args:
        image (np.ndarray): 2D image array.
        old_beam (Beam): Current beam.
        final_beam (Beam): Target beam.
        dx (u.Quantity): Pixel size in x.
        dy (u.Quantity): Pixel size in y.
        sfactor (float): Scaling factor.
        conbeam (Beam, optional): Convoling beam to use. Defaults to None.
        conv_mode (str, optional): Convolution mode to use. Defaults to "robust".

    Returns:
        np.ndarray: Smoothed image.
    """
    if np.isnan(sfactor):
        logger.warning("Beam larger than cutoff -- blanking")

        newim = np.ones_like(image) * np.nan
        return newim
    if conbeam is None:
        conbeam = final_beam.deconvolve(old_beam)
    if np.isnan(conbeam):
        return image * np.nan
    if np.isnan(image).all():
        return image
    if conbeam == Beam(major=0 * u.deg, minor=0 * u.deg, pa=0 * u.deg) and sfactor == 1:
        return image
    # using Beams package
    logger.debug(f"Old beam is {old_beam!r}")
    logger.debug(f"Using convolving beam {conbeam!r}")
    logger.debug(f"Target beam is {final_beam!r}")
    logger.debug(f"Using scaling factor {sfactor}")
    pix_scale = dy

    gauss_kern = conbeam.as_kernel(pix_scale)
    conbm1 = gauss_kern.array / gauss_kern.array.max()
    fac = sfactor
    if conv_mode == "robust":
        newim, fac = convolve(
            image.astype("f8"),
            old_beam,
            final_beam,
            dx,
            dy,
        )
        # keep the new sfactor computed by this method
        sfactor = fac
    if conv_mode == "scipy":
        newim = scipy.signal.convolve(image.astype("f8"), conbm1, mode="same")
    elif conv_mode == "astropy":
        newim = convolution.convolve(
            image.astype("f8"),
            conbm1,
            normalize_kernel=False,
        )
    elif conv_mode == "astropy_fft":
        newim = convolution.convolve_fft(
            image.astype("f8"),
            conbm1,
            normalize_kernel=False,
            allow_huge=True,
        )
    logger.debug(f"Using scaling factor {fac}")
    if np.any(np.isnan(newim)):
        logger.warning(f"{np.isnan(newim).sum()} NaNs present in smoothed output")

    newim *= fac

    # Ensure the output data-type is the same as the input
    return newim.astype(image.dtype)
